/*****************************************************************************
|*
|*  VERSION CONTROL:    @(#)monitor.c	1.13 99/10/21
|*
|*  IN PACKAGE:         monitor
|*
|*  COPYRIGHT:          Copyright (c) 1999 TASKING Software B.V.
|*
 ****************************************************************************/

#include "monitor.h"
#include "led.h"
#include "serial.h"
#include "sfr.h"
#include "rw.h"
#include "regbank.h"
#include "vector.h"

#if defined( _REGXASCC_SFR ) || defined( _REGXAH3_SFR ) || defined( _REGXAH4_SFR )
#include "dma.h"
#endif	/* defined( _REGXASCC_SFR ) || defined( _REGXAH3_SFR ) || defined( _REGXAH4_SFR ) */

_near unsigned char	        mode;		/* monitor mode */
static _near DATA 		data;		/* data buffer */
static _near ADDRESS 		address;	/* address buffer */
static _near unsigned char	space;		/* space selector */

/*****************************************************************************
|*
|*  FUNCTION:           transmit
|*
|*  AVAILABILITY:	static
|*
|*  PARAMETERS:
|*
|*  RETURN VALUE:
|*
|*  DESCRIPTION:
|*
|*	Transmit the low byte of the 32-bit DATA register to the host,
|*	right shifting it one byte.
|*
 */
static void _FMEM transmit( void )
{
	ser_out( data.b.b0 );
	data.b.b0 = data.b.b1;
	data.b.b1 = data.b.b2;
	data.b.b2 = data.b.b3;
}

/*****************************************************************************
|*
|*  FUNCTION:           receive
|*
|*  AVAILABILITY:	static
|*
|*  PARAMETERS:
|*
|*	input	byte received from host
|*
|*  RETURN VALUE:
|*
|*  DESCRIPTION:
|*
|*	Receive the low byte of the 32-bit DATA register from the host,
|*      left shifting it one byte.
|*
 */
static void _FMEM receive( unsigned char input )
{
	data.b.b3 = data.b.b2;
	data.b.b2 = data.b.b1;
	data.b.b1 = data.b.b0;
	data.b.b0 = input;
}

/*****************************************************************************
|*
|*  FUNCTION:           sfr
|*
|*  AVAILABILITY:	static
|*
|*  PARAMETERS:
|*
|*	c 	byte to write to SFR address location
|*	rw	0x00	read SFR address
|*		0x40	write SFR address
|*
|*  RETURN VALUE:
|*
|*	byte read from SFR address location
|*
|*  DESCRIPTION:
|*
|*      Patch code in data memory for accessing SFR's. Execute
|*	patched code by calling sfr_code overlayed with sfr_data.
|*
 */
static unsigned char _FMEM sfr( unsigned char c, unsigned char rw )
{
#ifdef RAM_SFR
	SFR_SPACE unsigned char volatile *	sfr_data = (SFR_SPACE unsigned char *)&sfr_code;
#endif	/* RAM_SFR */

	WRITE_CODE( &sfr_data[0], 0x86 );       /* MOV.B R0L,sfr_addr */
	WRITE_CODE( &sfr_data[1], (address.sfr.b1 & 0x7 | rw ) );
	WRITE_CODE( &sfr_data[2], address.sfr.b0 );
	WRITE_CODE( &sfr_data[3], 0xd6 );       /* RET */
	WRITE_CODE( &sfr_data[4], 0x80 );	
	
	c = sfr_code( c );                      /* execute patched code */

	return c;
}

/*****************************************************************************
|*
|*  FUNCTION:           reg_table
|*
|*  AVAILABILITY:	static
|*
|*  PARAMETERS:
|*
|*	byte	byte offset [0..3]
|*
|*  RETURN VALUE:
|*
|*	offset on stack register is pushed on
|*
|*  DESCRIPTION:
|*
|*      Get stack offset register is pushed on
|*
|*	Register numbering:
|*	R0H = 0, R0L = 1, R1H = 2, R1L = 3, R2H = 4, R2L = 5, R3H = 6, R3L = 7,
|*	R4H = 8, R4L = 9, R5H = 10, R5L = 11, R6H = 12, R6L = 13,
|*	R0 = 0, R1 = 2, R2 = 4, R3 = 6, R4 = 8, R5 = 10, R6 = 12
*/
static unsigned int _FMEM reg_table( unsigned int byte )
{
	static _ROM unsigned int table[14] = 
	{ 13+RTIFRAME, 14+RTIFRAME, 11+RTIFRAME, 12+RTIFRAME, 9+RTIFRAME, 
	  10+RTIFRAME,  7+RTIFRAME,  8+RTIFRAME,  5+RTIFRAME, 6+RTIFRAME, 
	  3+RTIFRAME,   4+RTIFRAME,  1+RTIFRAME,  2+RTIFRAME
	};

	return 	byte + table[ (address.reg & 0x1f) ];
}

/*****************************************************************************
|*
|*  FUNCTION:           read
|*
|*  AVAILABILITY:	static
|*
|*  PARAMETERS:
|*
|*	offset	stack offset parent function
|*	byte	byte offset [0..3]
|*
|*  RETURN VALUE:
|*
|*	byte read from active space
|*	
|*  DESCRIPTION:
|*
|*      Read byte from active space
|*
|*	All general purpose registers [R0..R6] are pushed on the stack,
|*	before the monitor is entered. Accessing user application registers
|*	is performed via the stack. The XA does not support indirect addressing
|*	of the SFR space. Instructions need to be patched in memory for direct
|*	SFR accces.
|*
 */
static unsigned char _FMEM read( unsigned int offset, unsigned int byte )
{
	unsigned char	c;
    
	switch ( space )			/* select active space */
	{
	case S_DATA:				/* data space */
		c = *(address.data + byte);	
		break;	
	case S_REG:					/* register space on stack */
		byte = reg_table( byte );		/* get byte register stack offset */
		c = *(BYTESP)USERSP( offset - RTIFRAME + 1 );	/* get user program PSWH */
		if ( (byte >= (7+RTIFRAME) )		/* byte is R0L/H..R3L/H */
		     && (c & 0x30) )			/* user program does not use monitor bank 0 */
		{
			c = regbank_read( c, byte );	/* read register 'byte' from bank 'c' */
		}
		else
		{
#if _MODEL == 'm' || _MODEL == 'c' || _MODEL == 'l'
	s_reg:
#endif	/* _MODEL == 'm' || _MODEL == 'c' || _MODEL == 'l' */
			if ( !( mode & MODE_RUN ) )     /* user appl. does not run ? */
			{
				byte += RTIFRAME;	/* add monitor interrupt frame */
			}
			c = *(BYTESP)USERSP( offset - byte );
		}
		break;
	case S_CODE:				/* code space */
		c = *(address.code + byte);
		break;
	case S_SFR:                             /* sfr space */
#if _MODEL == 'm' || _MODEL == 'c' || _MODEL == 'l'
		switch ( address.reg )
		{
		case 0x443:                     /* CS */
			byte = RTIFRAME + 16;   /* pushed on stack */
			goto s_reg;
		case 0x441:                     /* DS */
			byte = RTIFRAME + 18;   /* pushed on stack */
			goto s_reg;
		case 0x442:                     /* ES */
			byte = RTIFRAME + 20;   /* pushed on stack */
			goto s_reg;
		case 0x403:                     /* SSEL */
			byte = RTIFRAME + 22;   /* pushed on stack */
			goto s_reg;
		default:		
			c = sfr( 0, 0x00 );
			break;
		}
#else	/* _MODEL == 'm' || _MODEL == 'c' || _MODEL == 'l' */
		c = sfr( 0, 0x00 );
#endif	/* _MODEL == 'm' || _MODEL == 'c' || _MODEL == 'l' */
		break;
	default:                                /* illegal space */
		c = 0;
		break;
	}

	return c;    
}

/*****************************************************************************
|*
|*  FUNCTION:           read_byte
|*
|*  AVAILABILITY:	static
|*
|*  PARAMETERS:
|*
|*	offset	stack offset parent function
|*
|*  RETURN VALUE:
|*
|*  DESCRIPTION:
|*
|*	Read the byte of memory pointed to by the ADDR register and
|*	place it in the least significant byte of the DATA register
|*
 */
static void _FMEM read_byte( unsigned int offset )
{
	offset += OFFSP();                      /* add stack offset */
	data.b.b0 = read( offset, 0 );
}

/*****************************************************************************
|*
|*  FUNCTION:           read_word
|*
|*  AVAILABILITY:	static
|*
|*  PARAMETERS:
|*
|*	offset	stack offset parent function
|*
|*  RETURN VALUE:
|*
|*  DESCRIPTION:
|*
|*      Read the word of memory pointed to by the ADDR register and place it
|*	in the least significant word of the DATA register
|*
 */
static void _FMEM read_word( unsigned int offset )
{
	offset += OFFSP();                      /* add stack offset */
	if ( space == S_REG )
	{
		data.b.b0 = read( offset, 1 );
		data.b.b1 = read( offset, 0 );
	}
	else
	{
		data.b.b0 = read( offset, 0 );
		data.b.b1 = read( offset, 1 );
	}
}

/*****************************************************************************
|*
|*  FUNCTION:           read_long
|*
|*  AVAILABILITY:	static
|*
|*  PARAMETERS:
|*
|*	offset	stack offset parent function
|*
|*  RETURN VALUE:
|*
|*  DESCRIPTION:
|*
|*      Read the double-word of memory pointed to by the ADDR register and
|*	place it in the DATA register
|*
 */
static void _FMEM read_long( unsigned int offset )
{
	offset += OFFSP();                      /* add stack offset */
	if ( space == S_REG )
	{
		data.b.b0 = read( offset, 1 );
		data.b.b1 = read( offset, 0 );
		data.b.b2 = read( offset, 3 );
		data.b.b3 = read( offset, 2 );
	}
	else
	{
		data.b.b0 = read( offset, 0 );
		data.b.b1 = read( offset, 1 );
		data.b.b2 = read( offset, 2 );
		data.b.b3 = read( offset, 3 );
	}
}

/*****************************************************************************
|*
|*  FUNCTION:           write
|*
|*  AVAILABILITY:	static
|*
|*  PARAMETERS:
|*
|*	offset	stack offset parent function
|*      c       byte value to write
|*	byte	byte offset [0..3]
|*
|*  RETURN VALUE:
|*
|*	
|*  DESCRIPTION:
|*
|*	Write byte to active space, and increment the ADDR register
|*	by one.
|*
|*	To download code in the code space it need to be mapped on the
|*	data space. All general purpose registers [R0..R6] are pushed
|*	on the stack, before the monitor is entered. Modifing user
|*	application registers is performed via the stack. The XA does
|*	not support indirect addressing of the SFR space. Instructions
|*	need to be patched in memory for direct SFR accces.
|*
 */
static void _FMEM write( unsigned int offset, unsigned char c, unsigned int byte )
{
	unsigned char	pswh;

	switch ( space )			/* determine the destination space */
	{
	case S_CODE:				/* code mapped on data */
#ifdef RAM_CODE
	{
		DATAP	p = address.data + CODE_OFFSET + byte;

#ifdef	_USER_MON_VECTORS
		if ( cache_vectors( p, c ) )    /* check if ROM monitor vector is overwritten */
		{		
			break;
		}	
#endif	/* _USER_MON_VECTORS */
		if ( (unsigned long)p < IRAM_SIZE )	/* address in internal RAM  ? */
		{
			WRITE_CODE( (unsigned char _xdat *)p, c );	/* use MOVX */
		}
		else
		{
			WRITE_CODE( p, c );
		}
	}
#else	/* RAM_CODE */
#ifdef	_USER_MON_VECTORS
	{
		DATAP	p = address.data + CODE_OFFSET + byte;
	        
		if ( cache_vectors( p, c ) )	/* check if ROM monitor vector is overwritten */
		{
			break;
		}

		WRITE_CODE( p, c );
	}
#else	/* _USER_MON_VECTORS */
    	WRITE_CODE( address.data + CODE_OFFSET + byte, c );
#endif	/* _USER_MON_VECTORS */
#endif	/* RAM_CODE */
	break;
	case S_DATA:				/* data space */
#ifdef	_USER_MON_VECTORS
		if ( !cache_vectors( address.data + byte, c ) ) /* check if ROM monitor vector is overwritten */
#endif	/* _USER_MON_VECTORS */
		{		
			*(address.data + byte) = c;
		}
		break;	
	case S_REG:                             /* register space on stack */
		byte = reg_table( byte );	                	/* get byte register stack offset */
		pswh = *(BYTESP)USERSP( offset - RTIFRAME + 1 );	/* get user program PSWH */
		if ( (byte >= (7+RTIFRAME) )    &&              	/* byte is R0L/H..R3L/H */
		     (pswh & 0x30)              )               	/* user program does not use monitor bank 0 */
		{
			regbank_write( pswh, c, byte );	                /* write value 'c' in register 'byte' of bank 'pswh' */
		}
		else
		{
#if _MODEL == 'm' || _MODEL == 'c' || _MODEL == 'l'
	s_reg:
#endif	/* _MODEL == 'm' || _MODEL == 'c' || _MODEL == 'l' */
			if ( !( mode & MODE_RUN ) )     /* user appl. does not run ? */
			{
				byte += RTIFRAME;	/* add monitor interrupt frame */
			}
			*(BYTESP)USERSP( offset - byte ) = c;
		}
		break;
	case S_SFR:                             /* sfr space */	
#if _MODEL == 'm' || _MODEL == 'c' || _MODEL == 'l'
		switch ( address.reg )
		{
		case 0x443:				/* CS */
			byte = RTIFRAME + 16;		/* pushed on stack */
			goto s_reg;
		case 0x441:				/* DS */
			byte = RTIFRAME + 18;		/* pushed on stack */
			goto s_reg;
		case 0x442:				/* ES */
			byte = RTIFRAME + 20;		/* pushed on stack */
			goto s_reg;
		case 0x403:				/* SSEL */
			byte = RTIFRAME + 22;		/* pushed on stack */
			goto s_reg;
		default:		
			sfr( c, 0x8 );
			break;
		}
#else	/* _MODEL == 'm' || _MODEL == 'c' || _MODEL == 'l' */
		sfr( c, 0x8 );
#endif	/* _MODEL == 'm' || _MODEL == 'c' || _MODEL == 'l' */
		break;
	default:					/* illegal space */
		break;
	}
}

/*****************************************************************************
|*
|*  FUNCTION:           write_byte
|*
|*  AVAILABILITY:	static
|*
|*  PARAMETERS:
|*
|*	offset	stack offset parent function
|*
|*  RETURN VALUE:
|*
|*  DESCRIPTION:
|*
|*      Store the least significant byte of the DATA register in the byte of
|*	memory pointed to by the ADDR register, and increment the ADDR register
|*	by one, except for registers.
|*
 */
static void _FMEM write_byte( unsigned int offset )
{
	offset += OFFSP();                      /* add stack offset */
	write( offset, data.b.b0, 0 );
	if ( space != S_REG )
	{
		address.data++;
	}
}

/*****************************************************************************
|*
|*  FUNCTION:           write_word
|*
|*  AVAILABILITY:	static
|*
|*  PARAMETERS:
|*
|*	offset	stack offset parent function
|*
|*  RETURN VALUE:
|*
|*  DESCRIPTION:
|*
|*      Store the least significant word of the DATA register in the word of
|*	memory pointed to by the ADDR register, and increment the ADDR register
|*	by two, except for registers.
|*
 */
static void _FMEM write_word( unsigned int offset )
{
	offset += OFFSP();                      /* add stack offset */
	if ( space == S_REG )
	{
		write( offset, data.b.b0, 1 );
		write( offset, data.b.b1, 0 );
	}
	else
	{
		write( offset, data.b.b0, 0 );
		write( offset, data.b.b1, 1 );
		address.data += 2;
	}
}

/*****************************************************************************
|*
|*  FUNCTION:           write_long
|*
|*  AVAILABILITY:	static
|*
|*  PARAMETERS:
|*
|*	offset	stack offset parent function
|*
|*  RETURN VALUE:
|*
|*  DESCRIPTION:
|*
|*      Store the DATA register in the double-word of memory pointed to by
|*	the ADDR register, and increment the ADDR register by four, except
|*	for registers.
|*
 */
static void _FMEM write_long( unsigned int offset )
{
	offset += OFFSP();                      /* add stack offset */
	if ( space == S_REG )
	{
		write( offset, data.b.b0, 1 );
		write( offset, data.b.b1, 0 );
		write( offset, data.b.b2, 3 );
		write( offset, data.b.b3, 2 );
	}
	else
	{
		write( offset, data.b.b0, 0 );
		write( offset, data.b.b1, 1 );
		write( offset, data.b.b2, 2 );
		write( offset, data.b.b3, 3 );
		address.data += 4;
	}
}

/*****************************************************************************
|*
|*  FUNCTION:           read_psw
|*
|*  AVAILABILITY:	static
|*
|*  PARAMETERS:
|*
|*	offset	stack offset parent function
|*
|*  RETURN VALUE:
|*
|*  DESCRIPTION:
|*
|*      Load the low word of the DATA register with PSW
|*
 */
static void _FMEM read_psw( unsigned int offset )
{
	data.w.w0 = *(WORDSP)USERSP( offset - RTIFRAME );
}

/*****************************************************************************
|*
|*  FUNCTION:           write_psw
|*
|*  AVAILABILITY:	static
|*
|*  PARAMETERS:
|*
|*	offset	stack offset parent function
|*
|*  RETURN VALUE:
|*
|*  DESCRIPTION:
|*
|*      Load PSW with the low word of the DATA register
|*
 */
static void _FMEM write_psw( unsigned int offset )
{
	*(WORDSP)USERSP( offset - RTIFRAME ) = data.w.w0;
}

/*****************************************************************************
|*
|*  FUNCTION:           read_ssp
|*
|*  AVAILABILITY:	static
|*
|*  PARAMETERS:
|*
|*	offset	stack offset parent function
|*
|*  RETURN VALUE:
|*
|*  DESCRIPTION:
|*
|*      Load the DATA register with the SP (Stack Pointer) register
|*
 */
static void _FMEM read_ssp( unsigned int offset )
{
	data.w.w0 = USERSP( offset );
}

/*****************************************************************************
|*
|*  FUNCTION:           write_ssp
|*
|*  AVAILABILITY:	static
|*
|*  PARAMETERS:
|*
|*	offset	stack offset parent function
|*
|*  RETURN VALUE:
|*
|*  DESCRIPTION:
|*
|*      Load the SP (Stack Pointer) register with the DATA register
|*
|*	Copy stack up/down to set the user system stack pointer
|*
 */
static void _FMEM write_ssp( unsigned int offset )
{
	unsigned int	i;
	WORDSP		dst;
	WORDSP		src;

	offset += OFFSP();                      /* add stack offset */

	dst = (WORDSP)(data.w.w0 - offset);
	src = GETSP( 0 );

	if ( dst != src )
	{
		if ( dst < src )
		{
			for ( i = offset; i > 0; i -= 2 )
			{
				*dst++ = *src++;
			}
		}
		else
		{
			dst += offset/2;
			src += offset/2;

			for ( i = offset; i > 0; i -= 2 )
			{
				*--dst = *--src;
			}
		}
		SETSP( data.w.w0 - offset );
	}
}

/*****************************************************************************
|*
|*  FUNCTION:           read_pc
|*
|*  AVAILABILITY:	static
|*
|*  PARAMETERS:
|*
|*	offset	stack offset parent function
|*
|*  RETURN VALUE:
|*
|*  DESCRIPTION:
|*
|*      Load the DATA register with the user_PC (Program Counter) register
|*
 */
static void _FMEM read_pc( unsigned int offset )
{
#if _MODEL == 'm' || _MODEL == 'c' || _MODEL == 'l'
	data.w.w0 = *(WORDSP)USERSP( offset - (RETFRAME /2) );
	data.w.w1 = *(WORDSP)USERSP( offset - RETFRAME );
#else	/* _MODEL == 'm' || _MODEL == 'c' || _MODEL == 'l' */
	data.w.w0 = *(WORDSP)USERSP( offset - RETFRAME );    
	data.w.w1 = 0;
#endif	/* _MODEL == 'm' || _MODEL == 'c' || _MODEL == 'l' */
}

/*****************************************************************************
|*
|*  FUNCTION:           write_pc
|*
|*  AVAILABILITY:	static
|*
|*  PARAMETERS:
|*
|*	offset	stack offset parent function
|*
|*  RETURN VALUE:
|*
|*  DESCRIPTION:
|*
|*      Load the user_PC (Program Counter) register with the DATA register
|*
 */
static void _FMEM write_pc( unsigned int offset )
{
#if _MODEL == 'm' || _MODEL == 'c' || _MODEL == 'l'
	*(WORDSP)USERSP( offset - (RETFRAME / 2) ) = data.w.w0;
	*(WORDSP)USERSP( offset - RETFRAME ) = data.w.w1;
#else	/* _MODEL == 'm' || _MODEL == 'c' || _MODEL == 'l' */
	*(WORDSP)USERSP( offset - RETFRAME ) = data.w.w0;    
#endif	/* _MODEL == 'm' || _MODEL == 'c' || _MODEL == 'l' */
}

/*****************************************************************************
|*
|*  FUNCTION:           remove
|*
|*  AVAILABILITY:	static
|*
|*  PARAMETERS:
|*
|*	offset		stack offset parent function
|*
|*  RETURN VALUE:
|*
|*  DESCRIPTION:
|*
|*      Remove monitor interrupt frame from the stack to enter the user
|*	application.
|*
 */
static void _FMEM remove( unsigned int offset )
{
	unsigned int	i;
	WORDSP		dst;
	WORDSP		src;

	/*
	 *	Amount of bytes to copy is offset to user application subtracted
	 *	with the interrupt frame size of the user and monitor application.
	 */
	offset += OFFSP() - (RTIFRAME * 2);
    
	src = GETSP( offset );			/* get source stack pointer */
	dst = src + (RTIFRAME / 2);             /* get destination stack pointer */

	/*
	 *	This must be one continues block of code, because
	 *	stack pointer must be adjusted immediate after stack has been copied.
	 */
	{
		for ( i = offset; i > 0; i -= 2 )	/* move stack */
		{
			*--dst = *--src;
		}
		ADDSP( RTIFRAME );              /* monitor RTI frame removed */
	}
#ifdef	_USER_MON_VECTORS
	install_vectors();                      /* install all cached user vectors */
#endif	/* _USER_MON_VECTORS */
}

/*****************************************************************************
|*
|*  FUNCTION:           wait
|*
|*  AVAILABILITY:	static
|*
|*  PARAMETERS:
|*
|*  RETURN VALUE:
|*
|*  DESCRIPTION:
|*
|*      monitor wait loop
|*
 */
static void _FMEM wait( void )
{
	while ( 1 );
}

/*****************************************************************************
|*
|*  FUNCTION:           halt
|*
|*  AVAILABILITY:	static
|*
|*  PARAMETERS:
|*
|*	offset		stack offset parent function
|*
|*  RETURN VALUE:
|*
|*  DESCRIPTION:
|*
|*      Stop execution of user application and return to wait()
|*	The monitor frame is restored on the stack to leave the user application
|*
 */
static void _FMEM halt( unsigned int offset )
{
	unsigned int	i;
	WORDSP		src;
	WORDSP		dst;

	/*
	 *	Amount of bytes to copy is offset to user application subtracted
	 *	with the interrupt frame size of the user application.
	 */    
	offset += OFFSP() - RTIFRAME;  
    
	src = GETSP( 0 );                       /* get source stack pointer */
	dst = src - (RTIFRAME / 2);             /* get destination stack pointer */

	/*
	 *	This must be one continues block of code, because
	 *	stack pointer must be adjusted immediate after stack has been copied.
	 */
	{
		for ( i = offset; i > 0; i -= 2 )       /* move stack */
		{
			*dst++ = *src++;
		}
		ADDSP( -RTIFRAME );             /* monitor RTI frame added */
	}

	*dst++ = MON_PSW;                       /* set monitor psw */
	*(FUNCPTR *)dst = swap( wait );		/* set return address to monitor loop */

	mode &= ~MODE_RUN;                      /* signal user appl. not running */
}

/*****************************************************************************
|*
|*  FUNCTION:           commands
|*
|*  AVAILABILITY:	static
|*
|*  PARAMETERS:
|*
|*	command		monitor command
|*	offset		stack offset parent function
|*
|*  RETURN VALUE:
|*
|*  DESCRIPTION:
|*
|*      Process monitor commands
|*
 */
static void _FMEM commands( unsigned char command, unsigned int offset )
{	
	if ( !( mode & MODE_RUN ) )             /* user appl. does not run ? */
	{
		offset += RTIFRAME;             /* add monitor interrupt frame */
	}

	offset += OFFSP();                      /* add stack offset */
	
	switch ( command )
	{
		/*
		 *	Force the next byte received to be treated as data
		 *	instead of being treated as a command
		 */
	case C_SET_DLE:
		mode |= MODE_RECEIVE;
		break;
		/*
		 *	Loads the SELECTOR byte with the lowest byte of the DATA
		 *	register.Valid selector register values are: 00 for DATA space,
		 *	01 for registers, 02 for CODE space and 08 for SFR space
		 */
	case C_WRITE_SELECTOR:
		space = data.b.b0;
		break;
		/*
		 *	Same as the commands READ_BYTE and XMITA performed consecutively
		 */
	case C_READ_BYTE_XMITA:
		read_byte( offset );
		/* Fall through */
		/*
		 *	Same as Xmit, but also increment the ADDR register by one
		 */
	case C_XMITA:
		address.data++;
		/* Fall through */
	case C_XMIT:
		transmit();
		break;
	case C_READ_BYTE:
		read_byte( offset );
		break;
	case C_READ_WORD:
		read_word( offset );
		break;
	case C_READ_LONG:
		read_long( offset );
		break;	
	case C_WRITE_BYTE:
		write_byte( offset );
		break;
	case C_WRITE_WORD:
		write_word( offset );
		break;
	case C_WRITE_LONG:
		write_long( offset );
		break;
		/*
		 *	Load the ADDR register with the DATA register
		 */
	case C_DATA_TO_ADDR:
		address.data = (DATAP)data.l;
		break;
	case C_READ_PSW:
		read_psw( offset );
		break;
	case C_WRITE_PSW:
		if ( !( mode & MODE_RUN ) )     /* user appl. does not run ? */
		{
			write_psw( offset );
		}
		break;
	case C_READ_SP:
		read_ssp( offset );
		break;
	case C_WRITE_SP:
		if ( !( mode & MODE_RUN ) )     /* user appl. does not run ? */
		{
			write_ssp( offset );
		}
		break;
		/*
		 *	Load the DATA register with the USP (user stack pointer)
		 */
	case C_READ_USP:
		data.w.w0 = _getusp();
		break;
		/*
		 *	Load the USP (user stack pointer) with the DATA register
		 */
	case C_WRITE_USP:
		_setusp( data.w.w0 );
		break;
	case C_READ_PC:
		read_pc( offset );
		break;
	case C_WRITE_PC:
		if ( !( mode & MODE_RUN ) )     /* user appl. does not run ? */
		{
			write_pc( offset );
		}
		break;
		/*
		 *	Start execution of the user code at the address in the
		 *	user_PC register
		 */
	case C_GO:
		mode &= ~MODE_BREAK;
		if ( !( mode & MODE_RUN ) )     /* user appl. does not run ? */
		{
			mode |= MODE_RUN;       /* signal user appl. running */
			remove( offset );       /* enter user application */
#if defined( _REGXASCC_SFR ) || defined( _REGXAH3_SFR ) || defined( _REGXAH4_SFR )
			dma_on();               /* restore dma to user state */
#endif	/* defined( _REGXASCC_SFR ) || defined( _REGXAH3_SFR ) || defined( _REGXAH4_SFR ) */
		}
		break;
		/*
		 *	Step one instruction of user code
		 */
	case C_STEP:
		if ( !( mode & MODE_RUN ) )		/* user appl. does not run ? */
		{
			unsigned int save;
			unsigned int rtioff = 0;
		
			read_pc( offset );		/* get user PC */
			address.data = (DATAP)data.l;	/* load code pointer */
			space = S_CODE;			/* set code space */
			read_word( offset );		/* read instruction user appl. next executed */
			if ( data.w.w0 == 0x90d6 )	/* RETI instruction ? */
			{
				/*		 
				 *	It is not possible to step a single RETI instruction on the XA !
				 *	Always one extra instruction need to be executed in trace mode, because
				 *	the RETI instruction may not turn trace mode off by popping PSW from the stack. 
				 *	When trace mode is turned off in RETI instruction, undefined code execution
				 *	behavior of the XA core is detected ! This implementation only prevents the
				 *	monitor from crashing. When RETI step is performed the prolog of the
				 *	general interrupt handler is entered, just step through until user application
				 *	is entered.
				 */
				rtioff = RTIFRAME;	/* PSW of RETI instruction */
			}
			read_psw( offset + rtioff );	/* get user PSW */
			save = data.w.w0;		/* save user PSW */
			data.w.w0 |= 0x4F00;		/* set TM and highest IM in user appl. */
			/*
			 * NOTE: Setting the highest IM of user application is needed because the Trace
			 *	     mode of the XA does not work as described in 16-bit 80C51XA Microcontrollers
			 *	     Data Handbook. It seams that interrupt service routines with higher or
			 *	     equal priority are serviced before the instruction of user application is
			 *	     executed. The trace mode interrupt is then activated without having the
			 *	     instruction of the user application executed.
			 *
			 *       It is not possible to step any instruction changing PSWH bits
			 *       TM and IM bits !
			 */
			write_psw( offset + rtioff );	/* update user PSW */
			data.w.w0 = save;		/* restore user PSW */	    

			remove( offset );		/* enter user application */
		}
		break;
	case C_HALT:
		mode &= ~MODE_BREAK;
		if ( mode & MODE_RUN )			/* user application runs ? */
		{
#if defined( _REGXASCC_SFR ) || defined( _REGXAH3_SFR ) || defined( _REGXAH4_SFR )
			dma_off();			/* turn all dma off for XA-SCC */
#endif	/* defined( _REGXASCC_SFR ) || defined( _REGXAH3_SFR ) || defined( _REGXAH4_SFR ) */
			halt( offset );			/* return to wait() */
		}
		break;
		/*
		 *	Load the lowest byte of the DATA register with status
		 *	information, which is transmitted back to the host
		 *	(status values are: 00=stopped, 01=running, 02=trapped (at break)
		 */
	case C_REPORT:
		ser_out( mode & 0x03 );
		break;
		/*
		 *	Load the lowest byte of the DATA register with the
		 *	version of the monitor, which is transmitted back to
		 *	the host
		 */
	case C_REPORT_VERSION:
		ser_out( THIS_VERSION );
		break;
		/*
		 *	Force the next byte received to be treated as data
		 *	instead of being treated as a command
		 */	
	case C_DOWNLOAD:
		mode |= MODE_DOWNLOAD;
		break;
		/*
		 *	Data received from host
		 */
	default:
		receive( command );
		break;
	}
}

/*****************************************************************************
|*
|*  FUNCTION:           monitor
|*
|*  AVAILABILITY:	extern
|*
|*  PARAMETERS:
|*
|*	input	byte received
|*	offset	stack offset parent function
|*	
|*  RETURN VALUE:
|*
|*  DESCRIPTION:
|*
|*      Process monitor mode and determine the type of data byte received
|*
 */
extern void _FMEM monitor( unsigned char input, unsigned int offset )
{
	offset += OFFSP();                              /* add stack offset */
    
	if ( mode & MODE_DIAGNOSTIC )	        	/* diagnostic mode ? */
	{
		ser_out( input + BOUNCE_DIAGN );	/* transmit back */

		if ( input == DIAGN_CHAR1 || input == DIAGN_CHAR2 )
		{
			mode &= ~MODE_DIAGNOSTIC;	/* turn diagnostic mode off */
			led_on();			/* signal diagnostic mode off */
		}
	}
	else if ( mode & MODE_RECEIVE    )		/* receive data mode enabled ? */
	{
		mode &= ~MODE_RECEIVE;			/* disable data receive */
	
		receive( input );			/* store data in buffer */
	
		if ( mode & MODE_DOWNLOAD )		/* download data mode enabled ? */
		{
			write_byte( offset );	        /* store data in selected space */
		}
	}
	else if ( mode & MODE_DOWNLOAD )		/* download data mode enabled ? */
	{
		/*
		 *	check if input is a receive data mode enable,
		 *	for escaping a byte with download-command value.
		 */
		if ( input == C_SET_DLE )		/* receive command ? */
		{
			mode |= MODE_RECEIVE;		/* enable data receive */
		}
		/*
		 *	check if command is download data mode enable,
		 *	which will cause leaving download mode.
		 */
		else if ( input == C_DOWNLOAD )
		{
			mode &= ~MODE_DOWNLOAD;		/* disable download data mode */
		}
		else
		{
			receive( input );		/* store data in buffer */
			write_byte( offset );		/* store data in selected space */
		}
	}
	else			        		/* process monitor command */
	{
		commands( input, offset );
	}
}

/*****************************************************************************
|*
|*  FUNCTION:           stop
|*
|*  AVAILABILITY:	extern
|*
|*  PARAMETERS:
|*
|*  RETURN VALUE:
|*
|*  DESCRIPTION:
|*
|*      Break interrupt service routine
|*
 */
#if defined( _USER_MON_VECTORS ) && !defined( RAM_SFR )
INTERRUPT( -1, BREAK_PSW ) 
#else
INTERRUPT( 1, BREAK_PSW ) 
#endif
extern void _FMEM stop( void )
{
	mode |= MODE_BREAK;                     /* set user break bit */

	if ( mode & MODE_RUN )			/* user application runs ? */
	{
		/*
		 *	signal back to the host that we hit a breakpoint
		 */
		ser_out( MODE_WAIT );
	
		halt( OFFSP() );                /* return to wait() */
	}
}

/*****************************************************************************
|*
|*  FUNCTION:           trace
|*
|*  AVAILABILITY:	extern
|*
|*  PARAMETERS:
|*
|*  RETURN VALUE:
|*
|*  DESCRIPTION:
|*
|*      Trace interrupt service routine
|*
 */
#if defined( _USER_MON_VECTORS ) && !defined( RAM_SFR )
INTERRUPT( -1, TRACE_PSW ) 
#else	/* defined( _USER_MON_VECTORS ) && !defined( RAM_SFR ) */
INTERRUPT( 2, TRACE_PSW ) 
#endif	/* defined( _USER_MON_VECTORS ) && !defined( RAM_SFR ) */
extern void _FMEM trace( void )
{
	unsigned int	userIM = data.w.w0 & 0x0F00; /* get user mode IM bits */

	read_psw( OFFSP() );                    /* get user PSW */
	data.w.w0 &= 0xB0FF;			/* remove TM and highest IM in user appl. */
	data.w.w0 |= userIM;                    /* restore user IM bits */
	write_psw( OFFSP() );			/* update user PSW */

	halt( OFFSP() );                        /* return to wait() */
}

#if _MODEL == 'm' || _MODEL == 'c' || _MODEL == 'l'
/*****************************************************************************
|*
|*  FUNCTION:           swap
|*
|*  AVAILABILITY:	extern
|*
|*  PARAMETERS:
|*
|*	p	function pointer
|*
|*  RETURN VALUE:
|*
|*	swapped function pointer
|*
|*  DESCRIPTION:
|*
|*      Swap high and low word of a function pointer, to create valid return
|*	address to store on the stack. The XA stack grows from high to low so
|*	it has a "Big-Endian" word order on the stack, while in memory it has
|*	a "Little-Endian" word order !
|*
 */
extern FUNCPTR _FMEM swap ( FUNCPTR p )
{
	return (FUNCPTR)((((unsigned long)p) << 16) | (((unsigned long)p) >> 16));
}
#endif	/* _MODEL == 'm' || _MODEL == 'c' || _MODEL == 'l' */
